/************************************************************
 * libdconf is (C) CopyRight PTlink IRC Software 1999-2005  *
 *                 http://software.pt-link.net              *
 * This program is distributed under GNU Public License     *
 * Please read the file COPYING for copyright information.  *
 ************************************************************

  Description: main code for dconf parsing library

 *  $Id: dconf.c,v 1.1.1.1 2005/08/27 15:45:07 jpinto Exp $ 
*/

#include "stdinc.h"
#include "dconf.h"

ConfItem dconf_items[DCONF_MAX];

/* internal functions */
void dconf_log (const char *fmt, ...);
void dconf_del (char *name);

/*
 * dconf_add
 *
 * inputs       - item name
 *		- item type (CT_*)
 *		- item flags (CF_*)
 *		- item variable pointer
 *		- item default value
 * output       - item index
 * side effects - looks for a specific item name on the conf items list
 */
void
dconf_add (char *name, int type, int flags, void *vptr, char *def)
{
  int i = 0;
/*  dconf_log("Adding setting %s, %i (%s)", name, flags, def ? def : ""); */
  while (dconf_items[i].name && strcasecmp (dconf_items[i].name, name))
    ++i;
  if (dconf_items[i].name != NULL)	/* this item already exists */
    {				/* maybe check if same type ? */
    }
  else
    {
      dconf_items[i].name = strdup (name);
      dconf_items[i].type = type;
      dconf_items[i].vptr = vptr;
      dconf_items[i].flags = flags;
      dconf_items[i + 1].name = NULL;
    }

  if (def != NULL)
    dconf_change_item (name, def, CA_READ);
}

void dconf_del (char *name)
{
  int i = 0;

  /* lookup item */  
  while (dconf_items[i].name && strcasecmp (dconf_items[i].name, name))
    ++i;

  if (dconf_items[i].name != NULL)	/* this item exists */
    {
      free(dconf_items[i].name);
      if ((dconf_items[i].type == CT_STRING) && dconf_items[i].vptr)
        free( *(char**) dconf_items[i].vptr);
      
      while (dconf_items[++i].name != NULL)
	{
	  dconf_items[i - 1] = dconf_items[i];
	}
      dconf_items[i-1].name = NULL;
    }
}

/*
 * dconf_find_item
 *
 * inputs       - item name
 * output       - item index
 * side effects - looks for a specific item name on the conf items list
 */
int
dconf_find_item (char *name)
{
  int i = 0;

  while (dconf_items[i].name && strcasecmp (dconf_items[i].name, name))
    ++i;

  return dconf_items[i].name ? i : -1;
}

/*
 * dconf_check
 *
 * inputs       - none
 * output       - -1 if invalid, 0 otherwise
 * side effects - check if any required item is missing
 */

int
dconf_check (void)
{
  int i = 0;
  int errors = 0;
  int flags;
  while (dconf_items[i].name)
    {
      flags = dconf_items[i].flags;
      if ((flags & CF_REQ) && !(flags & CF_DEFINED))
	{
	  dconf_log ("Missing required directive: %s", dconf_items[i].name);
	  ++errors;
	}
      ++i;
    }

  if (errors)
    {
      dconf_log ("Terminating because of %i required item(s)", errors);
      return -1;
    }

  return 0;
}


/*
 * dconf_item_str
 *
 * inputs       - item index
 * output       - item value (as string)
 * side effects - returns a string with the item value
 */
char *
dconf_item_str (int i)
{
  static char tmpbuf[512];
  int tmpval;

  tmpbuf[0] = '0';

  if (!dconf_items[i].name)
    return tmpbuf;

  if (!(dconf_items[i].flags & CF_DEFINED))
    {
      strncpy (tmpbuf, "Undefined", sizeof(tmpbuf));
    }
  else
    {
      switch (dconf_items[i].type)
	{
	case CT_DESC:
	  break;

	case CT_STRING:
	  sprintf (tmpbuf, "\"%s\"", *(char **) dconf_items[i].vptr);
	  break;

	case CT_SET:
	  if (*(int *) dconf_items[i].vptr == -1)
	    sprintf (tmpbuf, "Yes");
	  else
	    sprintf (tmpbuf, "No");
	  break;

	case CT_TIME:
	  tmpval = *(int *) dconf_items[i].vptr;
	  if(tmpval > 86400)
	    sprintf (tmpbuf, "%dd", tmpval / 86400);
	  else if(tmpval > 3600)
	    sprintf (tmpbuf, "%dh", tmpval / 3600);
	  else if (tmpval > 59)
	    sprintf (tmpbuf, "%dm", tmpval / 60);
	  else
	    sprintf (tmpbuf, "%ds", tmpval);
	  break;

	case CT_INT:
	  tmpval = *(int *) dconf_items[i].vptr;
	  sprintf (tmpbuf, "%i", tmpval);
	  break;

	default:
	  strncpy (tmpbuf, "Unknown format", sizeof(tmpbuf));
	}

    }

  return tmpbuf;
}

void
dconf_remove_items (ConfItem citems[])
{
  int i = 0;
  while (citems[i].name != NULL)
    {
      dconf_del (citems[i++].name);
    }
}

/*
 * dconf_value()
 *
 * inputs       - item index
 * output       - item value (as string)
 * side effects - returns a string with the item value
 */
int
dconf_value (char *item)
{
  int i;
  i = dconf_find_item (item);

  if (i < 0)
    return 0;

  return *(int *) dconf_items[i].vptr;
}

/*
 * dconf_valuestr()
 *
 * inputs       - item index
 * output       - item value (as string)
 * side effects - returns a string with the item value
 */
char *
dconf_valuestr (char *item)
{
  int i;
  i = dconf_find_item (item);

  if (i < 0)
    return 0;

  return *(char **) dconf_items[i].vptr;
}
